/*
 * 3d.h in Virtual Combat
 * Copyright (c) 1993 by David Jameson
 *
 * Permission to use, copy, modify and distribute this software
 * for any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission appear in the supporting
 * documentation.
 */

/* Definitions for 3d routines */

#ifndef _3D_H_
#define _3D_H_

#define LOCALMODE 0   /* These are for the rotate_x, rotate_y and */
#define WORLDMODE 1   /* rotate_z routines. */

#define CLIPDIST 500  /* Distance at which objects are z-clipped */
#define MAXSIZEPOLY 32	/* Max number of sides a polygon can have */
#define MAXPOINTS 100  /* Max number of points an object can have */
#define CLXHEIGHT 1840	/* Height of player, used for collision detection */
#define CLXRADIUS /*300*/ 1200	 /* How close player can come to an object before
			   a collision is detected */
#define TABLESIZE 500  /* Max entries in sort-table */


/* Object flags (values for object.flags) */

#define PLAYER 1  /* This object represents a player */
#define ALWAYS_VISIBLE 2  /* Don't perform hidden surface removal */
#define CHECKX 4
#define CHECKY 8
#define CHECKZ 16
#define CHECKPOLY (CHECKX | CHECKY | CHECKZ)
#define CHECKOBJ 32
#define DRAW_FIRST 64 /* Draw object first, no matter how far away */
#define DRAW_LAST 128 /* Draw object last, no matter how far away */
/* DRAW_FIRST and DRAW_LAST have no effect if object is further
   away than its maximum distance; the object still won't be drawn */

/* The CHECK* flags are used to make depth sorting work a bit better.
   The primary depth sorting used is a simple quick-sort of the
   distance between the viewer and the centre point of each object.
   This will work with small objects far away from each other but there
   will be problems when a small object is close to a larger one.  To
   prevent anything strange from happening a test is done for objects
   which have CHECKOBJ set.  If the object is beside a large single
   polygon object then a test is done to see if the viewer and the object
   are on the same side of the object (e.g. if the polygon is horizontal
   then the test is done on the y coordinates).  Large single polygon
   objects should have CHECKX, CHECKY or CHECKZ set depending on which
   coordinates are to be checked. */


typedef struct {       /* Rotation matrix :				*/
    float a1, a2, a3;  /*   / newx \	   / a1  a2  a3 \    / oldx \	*/
    float b1, b2, b3;  /*  |  newy  |  =  |  b1  b2  b3  |  |  oldy  |	*/
    float c1, c2, c3;  /*   \ newz /	   \ c1  c2  c3 /    \ oldz /	*/
} matrix;


typedef struct {
    int x, y;  /* x and y coords of points on screen (relative to centre) */
} screenpoint;


typedef struct {
    long x, y, z;  /* Point in 3d space */
} point;


typedef struct {
    point position;	  /* Position of viewer */
    matrix rotmatrix;	  /* Rotation matrix of viewer */
    matrix inversematrix; /* Inverse rotation matrix of viewer */
    long zvel;		  /* Forward velocity of viewer */
    long yvel;		  /* Downward velocity of viewer */
    float xrot; 	  /* Rotation (radians) about the x-axis */
    float yrot; 	  /* Rotation (radians) about the y-axis */
} viewer_object;

typedef struct {
    color *col;   /* Pointer to base (max intensity) colour */
    float light;  /* Intensity of colour */
} surfcol;

typedef struct _object {
    int flags;
    point centre;      /* x, y and z coords of centre of object */
    matrix rotmatrix;  /* Rotation matrix */
    int no_points;     /* Number of points */
    point *points;     /* Pointer to point data */
    int no_surfaces;   /* Number of surfaces */
    unsigned short *surface;  /* Pointer to surface data */
    color *col;        /* Colour of object */
    float light;       /* Brightness of colour */
    long maxdist;      /* Maximum distance before object can't be seen */
    struct _object *farobj;    /* Pointer to less detailed object, drawn if
				  dist > maxdist (should be NULL if no other
				  object to be drawn) */
    short children;	       /* Number of child objects this object has */
    struct _object **childobj;	/* Pointer to array of child objects */
    struct _object *parentobj; /* Pointer to parent object */
    char *collide;     /* Pointer to a string which is sent to the
			  server if the player collides with this object */
    char *shoot;       /* Pointer to a string which is sent to the
			  server if the player shoots the object */
    char *portal;      /* Pointer to the name of the world into which the
			  player is teleported if he collides with the object */
    point teleport;    /* Position which player starts at after collision
			  with portal */
} object;

#define child_object(object, number) (object->childobj)[number]


typedef struct {
    float distance;
    object *objptr;
    point displ;     /* Displacement of object */
    point relative;  /* Displacement of object relative to viewer */
    matrix mat;      /* Orientation of object relative to viewer */
} sortentry;

extern viewer_object viewer;

/* Prototypes */

extern void polar_to_xyz(point *result, float length, float elev, float yrot);
extern void unit_matrix(matrix *mat);
extern void add_child(object *child, object *obj);
extern void delete_child(object *obj, int number);
extern void matmult(matrix *mat1, matrix *mat2, matrix *result);
extern void rotpoint(matrix *mat, point *pt);
extern void composite_matrix(object *obj, sortentry *entry);
extern void rotate_x(matrix *mat, float angle, int mode);
extern void rotate_y(matrix *mat, float angle, int mode);
extern void rotate_z(matrix *mat, float angle, int mode);
extern int rotate_points(sortentry *entry, point *pointdata);
extern void do_perspective(point *pointdata, screenpoint *screendata, int no_points);
extern void draw_object(point *rotpoints, screenpoint *screenpoints, object *obj,
    long *LOS_z, object **LOS_obj);
extern float distance(point *point1, point *point2);
extern int add_sort(object *obj, sortentry *table, int maxsize);
extern void depth_sort(sortentry *table, int size);
extern void draw_all(disp *d, sortentry *table, int size, long x, long y,
    long scale, object **LOS_obj, char **clxdata, char **portal,
    point *teleport);

extern void setup_hash(void);
extern void destroy(object *obj);
extern object *find_object(char *ob);
extern int parse(disp *d, char *script, point *start, int *startnum,
    object **wrld, object **players);

#endif
