/*
 * server.c in Virtual Combat
 * Copyright (c) 1993 by David Jameson
 *
 * Permission to use, copy, modify and distribute this software
 * for any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission appear in the supporting
 * documentation.
 */

/* Server program for multi-user 3D environment */

#include "defs.h"

#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/telnet.h>

/* telnet.h on SunOS doesn't define some of the telnet options */

#ifndef TELOPT_LINEMODE
#define TELOPT_LINEMODE 34
#endif

#include <limits.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include "protocol.h"

/*
 * Undefine CHECK_IP to allow more than one player to log in from the
 * same IP address.
 */

#define CHECK_IP
#define SALT "Ag"               /* Salt used for crypt() */
#define WORLDFILE "world"       /* 3d object descriptions for world */
#define WORLDDEFS "world.defs"  /* Script file common to all worlds */
#define SAVEFILE "savefile"     /* File used for saving scores */
#define TEMPFILE "savefile.temp"
#define TELNET_PORT 2003	/* Port for telnet connections */
#define LINES_ON_SCREEN 24	/* Lines of text drawn using more() */
#define WARN 10 		/* Warn player of energy loss when each time
				   this amount is lost */
#define HEAL_ENERGY (MAX_ENERGY / 5)   /* Energy gained by "healing object" */
#define BLAST_DAMAGE (MAX_ENERGY / 5)  /* Energy lost from grenade blast */
#define HEAL_CHANCE 2000	 /* Chance of a heal_object appearing */
#define HEAL_DESTROY_CHANCE 800  /* Chance of a heal_object disappearing */
#define HIGH_SCORES 20	  /* Size of high score table */

/*
 * If during a game no players lose any energy within MAX_IDLE_TIME
 * seconds then the game will automatically stop.
 * Undefine this constant to disable this time check.
 * Similarly, MAX_WAIT defines the maximum time to wait for all clients
 * to send their data to the server (during a game).
 */

#define MAX_IDLE_TIME 300

/*
 * If no input arrives from any player (text or game data) for
 * MAX_PLAYER_IDLE seconds then the player will be removed.
 */

#define MAX_PLAYER_IDLE 1200
#define MAX_WAIT 10
#define ENDLIST 10000	  /* Marker to indicate end of linked list */
#define IPLEN 20	  /* Max length of IP address */
#define PROMPTLEN 20	  /* Max size of prompt */
#define TYPELEN 20	  /* Max length of game version */
#define PASSLEN 13	  /* Length of DES encrypted password */
#define FILENAMELEN NAMELEN /* Length of world filenames */
#define HELPDIR "helpdir" /* Where the help files are stored */
#define CLIENT 1	  /* Player connected using normal client program */
#define LOCAL 2 	  /* Connected via stdin & stdout */

#define ETHER_HDR 14
#define IP_HDR 20
#define TCP_HDR 20
#define PKTSIZE_AVERAGE (ETHER_HDR + IP_HDR + TCP_HDR + HDRSIZE + \
			 sizeof (update))
#define ETHER_SPEED 10000000	 /* 10Mb ethernet */

/* Comment out the next line to disable players connecting using telnet */
#define TELNET 8	  /* Player connected using TELNET */

typedef struct _player {
    int sckt;		/* Socket for this player */
    int mode;		/* GAMEMODE, TEXTMODE etc. */
    int connect;	/* Method of connection */
    int energy; 	/* Reduced each time player is shot */
    FILE *more_file;	/* File being viewed using more() */
    int more_file_position;	/* Current position in file */
    void (*func)();     /* If this is not NULL then this function will
			   be called every time a player enters a command */
    int saved_scores;	/* Have scores been saved since they last changed? */
    char name[NAMELEN + 1];	 /* Player's name */
    int idle_time;		 /* Idle time, in seconds */
    long login_time;		 /* Time of login */
    char world[FILENAMELEN + 1]; /* Name of world which player is in */
    char ip[IPLEN + 1]; 	 /* Domain name or IP address of player */
    char version[TYPELEN + 1];	 /* Version number of client */
    char prompt[PROMPTLEN + 1];  /* Prompt given after each command */
    char password[PASSLEN + 1];  /* Encrypted password */
    int games;	      /* Number of games played */
    int wins;	      /* Number of games won */
    int kills;	      /* Number of kills */
    unsigned char colour;  /* Index into playercols[] array */
    int nexttext;     /* Pointer to next text player in linked list (points to
			  array position, not memory location) */
    int nextgame;     /* Next player in gameplayers linked list */
    int nextQ;	      /* Next player in the queue */
} player;

typedef struct {
    char name[NAMELEN + 1];
    int kills;
    int games;
    int wins;
} high_score;


/* prototypes */

char *crypt(char *key, char *salt);

static void check_energy(int id, int plr);
static void usage(char *name);
static void check_heal(void);
static void doio(int id);
static int strip_telnet_commands(unsigned char *to, unsigned char *from,
    int bytes, int id);
static void parse_input(int id);
static long get_time(void);
static void leave_gamemode(int id);
static void enter_gamemode(int id);
static void chat(int id);
static void more(int id);
static void more_file(int id, char *name);
static int purge(int months);
static int check_colour(int check_colour);
static void parse_data_all(unsigned char *data);
static void parse_data(int id, unsigned char *data);
static void parse_file_all(char *name);
static void parse_file(int id, char *name);
static void make_connection_telnet(void);
static void make_connection(void);
static void check_high_score(player *plr);
static int read_high_scores(void);
static void get_name(int id);
static void get_password(int id);
static void change_password(int id);
static void change_password2(int id);
static void change_password3(int id);
static void new_password(int id);
static void new_password2(int id);
static void echo_on(int id);
static void echo_off(int id);
static void parse_data_all_world(unsigned char *data, char *world);
static void setup_socket(void);
static void sighandler(int sig);
static void remove_all_players(void);
static void remove_player(int id);
static void alrm(int sig);
static void sendtxt_all(unsigned char *text);
static void sendtxt(int id, unsigned char *text);
static void sendall(int mode, databuffer *data);
static int delQ(int id);
static int addQ(int id);
static void delgame(int id);
static void deltext(int id);
static void addgame(int id);
static void addtext(int id);
static int find_player(char *name);
static void cat_file(int id, char *name);
static int senddata(int id, databuffer *data);
static int write_player_data(int id);
static int read_player_data(int id);
static int write_player_entry(FILE *file, player *plr);
static int read_player_entry(FILE *file, player *plr);
static int stricmp(char *str1, char *str2);
static char *strcat2(char *str1, char *str2);
static databuffer *allocbuf(int size);



/* input_to() will reroute a player's input to a given function.  The
   function will be called with a single argument: the player's id */
#define input_to(id, function) (players[id].func = (void (*)())function)

static char *wspace = " \n\t\v\r\f";  /* White space characters */
static char version[] = "qub_1.0";  /* Version number */
static player players[MAX_LOGINS];
static high_score high_scores[HIGH_SCORES];
static int logins, numtext, numgame;
static int textplayers, gameplayers;  /* Root nodes for the two linked lists */
static int queue;		  /* Queue for people waiting to play game */
static int allow_games; 	  /* Allow players to enter game mode ? */
static int listen_socket, listen_socket2;
static long drops;		  /* Dropped packets */
static long inpackets;		  /* Total packets received */
static long outpackets; 	  /* Total packets sent */
static long up_since;		  /* Time when server was started */
static long games;		  /* Games played since server started */
static databuffer inbuffer;	  /* Buffer for incoming data */
static databuffer outbuffer;	  /* Buffer for outgoing data */
static databuffer gamebuffer;	  /* Buffer for outgoing game data */
static int gamecount;		  /* Number of clients which have sent game data
				     during this frame */
static unsigned char *gamebufptr; /* Pointer to current position in gamebuffer */
static int frames;
static int idle_time;		  /* Idle time in seconds */
static int wait_time;		  /* How long we've been waiting for
                                     game data in seconds */
static int framespersec;
static int heal_object; 	  /* Does a heal_object exist? */
static char default_prompt[] = "$";
static int server_id;		  /* ID of user logged in at server (negative
				     for no login) */
static int access_on;		  /* Set to indicate that new connections are being
				     accepted */
static int shutdownmode;	  /* Set to indicate shutdown taking place */
static int time_to_start;	  /* Length of time until game starts */

static char *playercols[] = { "blue", "green", "yellow", "red", "brown",
			      "pink", "orange", "violet" };


/* Allocate a databuffer structure with space for <size> bytes of data */
static databuffer *allocbuf(int size)
{
    return (databuffer *)malloc(HDRSIZE + size);
}  /* allocbuf() */


/* Concatenates str2 onto str1 and returns str1.  If str1 is NULL it
   concatenates onto the end of the last string passed (i.e. more
   efficient). */
static char *strcat2(char *str1, char *str2)
{
    static char *prev, *start;

    if (! str1) str1 = prev;
    start = str1;
    while (*str1 != '\0') str1++;
    while (*str2 != '\0') *str1++ = *str2++;
    *str1 = '\0';
    prev = str1;  /* Next time it won't be necessary to find the end of
		     str1 */
    return start;
}  /* strcat2() */


/* Case insensitive string comparison.	Returns 0 if strings are the same. */
static int stricmp(char *str1, char *str2)
{
    if (*str1 == '\0' && *str2 == '\0') return 0;
    if (*str1 == '\0' || *str2 == '\0') return 1;
    do {
	if (toupper(*str1++) != toupper(*str2++)) return 1;
    } while (*str1 != '\0' && *str2 != '\0');
    if (*str1 == '\0' && *str2 == '\0') return 0;
    return 1;
}  /* stricmp() */


/* Turn current time into a long integer */
static long get_time(void)
{
    time_t t;
    struct tm *ltime;
    long retval;

    if (! (t = time(NULL))) {
	perror("time()");
	return 0;
    }

    ltime = localtime(&t);

    return  (((((((ltime->tm_year << 4) + ltime->tm_mon) << 5) +
	      ltime->tm_mday) << 5) + ltime->tm_hour) << 6) + ltime->tm_min;
}  /* get_time() */


/* Prints the time into a buffer which should contain at least 25 bytes */
static void print_time(long time_value, char *buffer)
{
    static char *months[] = {"January", "February", "March", "April", "May",
    	"June", "July", "August", "September", "October", "November",
	"December" };
    int year, month, day, hour, minute;

    year = time_value >> 20;	/* years since 1900 */
    month = (time_value >> 16) & 0xf;
    assert (month >= 0 && month <= 11);
    day = (time_value >> 11) & 0x1f;
    hour = (time_value >> 6) & 0x1f;
    minute = time_value & 0x3f;
	
    sprintf(buffer, "%d:%d, %d %s %d", hour, minute, day, months[month],
    	year + 1900);
} /* print_time() */


/* Read an entry from the save-game file.  Returns FALSE on failure. */
static int read_player_entry(FILE *file, player *plr)
{
    int hackmode;

    if (fscanf(file, "%20[^:]:%13[^:]:%d:%d:%d:%d:%ld:%20[^\n]\n",
	plr->name, plr->password, &hackmode, &plr->games, &plr->wins,
	&plr->kills, &plr->login_time, plr->ip) != 8)
	return FALSE;  /* failure */
    if (hackmode) plr->mode = HACKMODE; else plr->mode = 0;
    return TRUE;  /* success */
}  /* read_player_entry() */


/* Write an entry to the save-game file.  Returns FALSE on failure. */
static int write_player_entry(FILE *file, player *plr)
{
    fprintf(file, "%s:%s:%d:%d:%d:%d:%ld:%s\n", plr->name, plr->password,
      (plr->mode & HACKMODE) ? 1 : 0, plr->games, plr->wins, plr->kills,
       plr->login_time, plr->ip);
    return TRUE;  /* success */
}  /* write_player_entry() */


/*
 * Read player information from file and create file if it doesn't exist.
 * Returns FALSE on failure.
 */
static int read_player_data(int id)
{
    FILE *savefile;
    player temp;

    if (! (savefile = fopen(SAVEFILE, "a"))) {  /* Create file if it doesn't
						   exist */
	perror("fopen()");
	return FALSE;
    }
    fclose(savefile);
    if (! (savefile = fopen(SAVEFILE, "r"))) {
	perror("fopen()");
	return FALSE;
    }
    while (! feof(savefile)) {
	if (! read_player_entry(savefile, &temp)) {
	    if (feof(savefile)) break;
	    fprintf(stderr, "Error in format of save-file.\n");
	    fclose(savefile);
	    return FALSE;
	}
	if (stricmp(players[id].name, temp.name) == 0) {
	    strcpy(players[id].password, temp.password);
	    players[id].mode |= temp.mode;
	    players[id].games = temp.games;
	    players[id].wins = temp.wins;
	    players[id].kills = temp.kills;
	    fclose(savefile);
	    return TRUE;
	}
    }
    fclose(savefile);
    return FALSE;
}  /* read_player_data() */


/* Write player information to file.  Returns FALSE on failure. */
static int write_player_data(int id)
{
    FILE *savefile, *tempfile;
    player temp;
    int found;

    found = FALSE;  /* player hasn't been found in the file */
    if (! (savefile = fopen(SAVEFILE, "r"))) {
	perror("fopen()");
	return FALSE;
    }
    if (! (tempfile = fopen(TEMPFILE, "w"))) {
	perror("fopen()");
	return FALSE;
    }
    while (! feof(savefile)) {
	if (! read_player_entry(savefile, &temp)) {
	    if (feof(savefile)) break;
	    fprintf(stderr, "Error in format of savefile.\n");
	    fclose(savefile);
	    fclose(tempfile);
	    remove(TEMPFILE);
	    return FALSE;
	}
	if (stricmp(players[id].name, temp.name) == 0) {
	    found = TRUE;
	    if (! write_player_entry(tempfile, &players[id])) {
		perror("fprintf()");
		fclose(savefile);
		fclose(tempfile);
		remove(TEMPFILE);
		return FALSE;
	    }
	}
	else {
	    if (! write_player_entry(tempfile, &temp)) {
		perror("fprintf()");
		fclose(savefile);
		fclose(tempfile);
		remove(TEMPFILE);
		return FALSE;
	    }
	}
    }
    if (! found) {
	if (! write_player_entry(tempfile, &players[id])) {
	    perror("fprintf()");
	    fclose(savefile);
	    fclose(tempfile);
	    remove(TEMPFILE);
	    return FALSE;
	}
    }

    if (ferror(tempfile)) {
	sendtxt_all("Corrupted savefile: shutting down server.\n");
	perror("bad savefile");
	remove_all_players();
	exit(1);
    }

    fclose(savefile);
    fclose(tempfile);

    if (rename(TEMPFILE, SAVEFILE)) {
	sendtxt_all("Corrupted savefile: shutting down server.\n");
        perror("rename()");
        remove_all_players();
	exit(1);
    }

    players[id].saved_scores = TRUE;
    return TRUE;
}  /* write_player_data() */


/* Checks whether a player has moved up in the high score table (or into it) */
static void check_high_score(player *plr)
{
    int i;
    int position;   /* Position of player in table (or zero) */

    position = 0;

    if (plr->kills > high_scores[0].kills) {
	for (i = 0; i < HIGH_SCORES; i++) {
	    if (! stricmp(plr->name, high_scores[i].name)) {
		position = i;
		break;
	    }
	}

	/* Assume players only move up in the table */

	for (i = position;
	  plr->kills > high_scores[i + 1].kills && i < HIGH_SCORES - 1;
	  i++) {
	    high_scores[i] = high_scores[i + 1];
	}
	strcpy(high_scores[i].name, plr->name);
	high_scores[i].kills = plr->kills;
	high_scores[i].games = plr->games;
	high_scores[i].wins = plr->wins;
    }
}  /* check_high_score() */


/*
 * Generate high score table by reading from the savefile.
 * Creates savefile if it doesn't already exist.
 * Returns FALSE on failure.
 */
static int read_high_scores(void)
{
    FILE *savefile;
    player temp;
    int i;

    for (i = 0; i < HIGH_SCORES; i++) {
	high_scores[i].name[0] = '\0';
	high_scores[i].kills = high_scores[i].wins = high_scores[i].games = 0;
    }

    if (! (savefile = fopen(SAVEFILE, "a"))) {  /* Create file if it doesn't
						   exist */
	perror("fopen()");
	return FALSE;
    }
    fclose(savefile);

    if (! (savefile = fopen(SAVEFILE, "r"))) {
	perror("fopen()");
	return FALSE;
    }
    while (! feof(savefile)) {
	if (! read_player_entry(savefile, &temp)) {
	    if (feof(savefile)) break;
	    fprintf(stderr, "Error in format of save-file.\n");
	    fclose(savefile);
	    return FALSE;
	}
	check_high_score(&temp);
    }
    fclose(savefile);
    return TRUE;
}  /* read_high_scores() */


/*
 * Removes all players who haven't logged in within a certain number
 * of months.
 * Returns FALSE if an error occurred.
 */
static int purge(int months)
{
    long check_time;
    FILE *savefile, *tempfile;
    player temp;
    int year, month, day;

    check_time = get_time();
    year = check_time >> 20;
    month = (check_time >> 16) & 0xf;
    day = (check_time >> 11) & 0x1f;

    month -= months;
    while (month < 0) {
	month += 12;
	year -= 1;
    }

    check_time = (year << 20) + (month << 16) + (day << 11);
    fprintf(stderr, "Purging all players not logged in since %d/%d/%d\n",
	    day, month + 1, year + 1900);

    if (! (savefile = fopen(SAVEFILE, "r"))) {
	perror("fopen()");
	return FALSE;
    }
    if (! (tempfile = fopen(TEMPFILE, "w"))) {
	perror("fopen()");
	return FALSE;
    }
    while (! feof(savefile)) {
	if (! read_player_entry(savefile, &temp)) {
	    if (feof(savefile)) break;
	    fprintf(stderr, "Error in format of savefile.\n");
	    fclose(savefile);
	    fclose(tempfile);
	    remove(TEMPFILE);
	    return FALSE;
	}
	if (temp.login_time > check_time) {
	    /*
	     * Player has logged on since "check_time", so write him
	     * back out to the savefile.
	     */
	    if (! write_player_entry(tempfile, &temp)) {
		perror("fprintf()");
		fclose(savefile);
		fclose(tempfile);
		remove(TEMPFILE);
		return FALSE;
	    }
	}
	else fprintf(stderr, "%s purged\n", temp.name);
    }
    fclose(savefile);
    fclose(tempfile);
    if (rename(TEMPFILE, SAVEFILE)) {
       perror("rename()");
       remove(TEMPFILE);
       return FALSE;
    }
    return TRUE;
}  /* purge() */



/* Send data to a player.  Returns FALSE on failure. */
static int senddata(int id, databuffer *data)
{
    int length, sent;
    int nds;  /* Max number of file descriptors (for select() ) */
    fd_set writefds;
    int num;
    struct timeval timeout;

    nds = getdtablesize();

    if (players[id].mode == NOTINUSE) return FALSE;

    /* Wait a maximum of 5 seconds for socket to become ready for writing */

    timeout.tv_sec = 5;
    timeout.tv_usec = 0;

    length = ntohs(data->length);
    ++outpackets;
    do {
	FD_ZERO(&writefds);
	FD_SET(players[id].sckt, &writefds);

	num = select(nds, NULL, &writefds, NULL, &timeout);

	if (num < 0) {
	    perror("select()");
	    remove_player(id);
	    return;
	}
	if (num != 1) {
	    fprintf(stderr, "select() returns %d\n", num);
	}

	sent = write(players[id].sckt, (char *)data, length);
	if (sent == 0) {
#ifdef DEBUG
	    fprintf(stderr, "Tried to send data to closed socket\n");
#endif
	    return FALSE;
	}
	if (sent == -1) {
	    perror("write()");
	    remove_player(id);
	    ++drops;
	    return FALSE;
	}
	length -= sent;
	data = (databuffer *)(sent + (char *)data);
#ifdef DEBUG
	if (length) fprintf(stderr, "write(): still %d bytes to write\n",
			    length);
#endif
	if (length < 0) {
	    /* We may as well check */
	    fprintf(stderr, "senddata(): negative length!\n");
	    exit(1);
	}
    } while (length);
    return TRUE;
}  /* senddata() */


/* List a text file */
static void cat_file(int id, char *name)
{
    struct stat buf;
    databuffer *sendbuf;
    FILE *infile;

    if (! (infile = fopen(name, "r"))) {
	perror("fopen()");
	return;
    }
    if (stat(name, &buf)) {
	perror("stat()");
	exit(1);
    }
    if (buf.st_size > DATASIZE - 1) {
	fprintf(stderr, "File too long.\n");
	return;
    }
    if (! (sendbuf = allocbuf(buf.st_size + 1))) {
	perror("malloc()");
	exit(1);
    }
    if (fread(sendbuf->data, buf.st_size, 1, infile) != 1) {
	 perror("fread()");
	 free(sendbuf);
	 return;
    }
    fclose(infile);
    sendbuf->data[buf.st_size] = '\0';  /* NULL terminator */
    if (players[id].connect & LOCAL) {
	fputs(sendbuf->data, stdout);
	fflush(stdout);
    }
#ifdef TELNET
    else if (players[id].connect & TELNET) {
	int length, sent;

	length = buf.st_size + 1;
	while (length > 0) {
	    sent = write(players[id].sckt,
			 sendbuf->data + buf.st_size + 1 - length, length);
	    if (sent == -1) {
		++drops;
		perror("write()");
		break;
	    }
	    length -= sent;
#ifdef DEBUG
	    if (length) fprintf(stderr, "write(): still %d bytes to write\n",
				length);
#endif
	}
	++outpackets;
    }
#endif
    else {
	sendbuf->length = htons(buf.st_size + 1 + HDRSIZE);
	sendbuf->pkttype = TEXTDATA;
	(void)senddata(id, sendbuf);
    }
    free(sendbuf);
}  /* cat_file() */


static void more(int id)
{
    char buf[LINES_ON_SCREEN * 81 + 1];     /* 80 char lines + linefeeds */
    int columns, lines;
    int character;
    char *bufptr;
    FILE *file;

    lines = 0;
    bufptr = buf;
    file = players[id].more_file;

    if (inbuffer.data[0] == 'q') {
	if (fclose(players[id].more_file)) perror("fclose()");
	input_to(id, NULL);
	return;
    }

    /* Set position in file to position of the end of the last read */

    if (fseek(file, players[id].more_file_position, SEEK_SET)) {
	perror("fseek()");
	input_to(id, NULL);
	return;
    }

    do {
	columns = 0;
	do {
	    character = getc(file);
	    if (feof(file)) break;
	    if (character == '\n') break;
	    *bufptr++ = character;
	    columns++;
	}  while (columns < 80);
	*bufptr++ = '\n';
	if (feof(file)) break;
	lines++;
    } while (lines < LINES_ON_SCREEN - 1);

    *bufptr++ = '\0';

    if (! feof(file)) strcat(buf, "Press <RETURN> for more or 'q' to quit:");
    sendtxt(id, buf);

    if (feof(file)) {
	if (fclose(file)) perror("fclose()");
	input_to(id, NULL);
	return;
    }

    /* Set up position for next time */

    players[id].more_file_position = ftell(file);
}  /* more() */


/* List a file to a player, a page at a time */
static void more_file(int id, char *name)
{
    char buf[100];

    if (! (players[id].more_file = fopen(name, "r"))) {
	sprintf(buf, "Can't open file %s.\n", name);
	sendtxt(id, buf);
	return;
    }

    /* Place initial position in file in the player's more_file_position */

    players[id].more_file_position = ftell(players[id].more_file);

    input_to(id, more);     /* Called each time player presses <RETURN> */
    more(id);   /* Send first part of file */
}  /* more_file() */


/* Find a player's id, given their name */
static int find_player(char *name)
{
    int id;

    id = 0;
    while ((stricmp(players[id].name, name)) && (id < MAX_LOGINS - 1)) id++;
    if (! stricmp(players[id].name, name)) return id;
    else return -1;  /* Error */
}  /* find_player() */


/* Check to see if this player is logged in twice */
static void check_ip(int id)
{
    int i;

    /* No need to check players who aren't connected using a client */

    if (! (players[id].connect & CLIENT)) return;

    if (players[id].mode & HACKMODE) return;

    for (i = 0; i < MAX_LOGINS; i++) {
	if ((players[i].mode != NOTINUSE) && (players[i].connect & CLIENT)
	  && (strcmp(players[id].ip, players[i].ip) == 0) && (i != id)
	  && (! (players[i].mode & HACKMODE))) {
	    fprintf(stderr,
	      "Players %s and %s logged on from the same address.\n",
	      players[id].name, players[i].name);
	    sendtxt(id, "A player is already logged in from your machine!\n");  
	    remove_player(id);  
	    return;
	}
    }
}  /* check_ip() */


/* Add a player to textplayers linked list */
static void addtext(int id)
{
    /* Just check to see if player is already at the start of the linked
       list because this could cause problems */
    if (textplayers == id) {
	fprintf(stderr, "Error: player already in linked list.\n");
	exit(1);
    }
    players[id].nexttext = textplayers;
    textplayers = id;
}  /* addtext() */


/* Add a player to gameplayers linked list */
static void addgame(int id)
{
    if (gameplayers == id) {
	fprintf(stderr, "Error: player already in linked list.\n");
	exit(1);
    }
    players[id].nextgame = gameplayers;
    gameplayers = id;
}  /* addgame() */


/* Delete a player from textplayers linked list */
static void deltext(int id)
{
    int ptr;

    ptr = textplayers;
    if (ptr == ENDLIST) return;
    if (ptr == id) textplayers = players[id].nexttext;
    else {
	while ((players[ptr].nexttext != ENDLIST) &&
	       (players[ptr].nexttext != id))
	    ptr = players[ptr].nexttext;
	if (players[ptr].nexttext == id)
	    players[ptr].nexttext = players[id].nexttext;
    }
}  /* deltext() */


/* Delete a player from gameplayers linked list */
static void delgame(int id)
{
    int ptr;

    ptr = gameplayers;
    if (ptr == ENDLIST) return;
    if (ptr == id) gameplayers = players[id].nextgame;
    else {
	while ((players[ptr].nextgame != ENDLIST) &&
	       (players[ptr].nextgame != id))
	    ptr = players[ptr].nextgame;
	if (players[ptr].nextgame == id)
	    players[ptr].nextgame = players[id].nextgame;
    }
}  /* delgame() */


/* Add player to queue of people waiting to play. Returns position in queue
   or 0 if player already in queue */
static int addQ(int id)
{
    int Qptr, lastptr;
    int posn;  /* position in Queue */

    /* Add to end of queue */
    posn = 1;
    if (queue == ENDLIST) {
	queue = id;
	players[id].nextQ = ENDLIST;
    }
    else {
	lastptr = Qptr = queue;
	while (Qptr != ENDLIST) {
	    posn++;
	    if (Qptr == id) return 0;  /* already in queue */
	    lastptr = Qptr;
	    Qptr = players[Qptr].nextQ;
	}
	players[id].nextQ = ENDLIST;
	players[lastptr].nextQ = id;
    }
    return posn;
}  /* addQ() */

/*
 * Delete a player from anywhere in the queue.	Returns TRUE if player
 * was in the queue (otherwise does nothing and returns FALSE).
 */
static int delQ(int id)
{
    int Qptr, lastptr;

    if (queue == ENDLIST) return FALSE;
    if (queue == id) {
	queue = players[id].nextQ;
	return TRUE;
    }
    lastptr = Qptr = queue;
    while (Qptr != id && Qptr != ENDLIST) {
	lastptr = Qptr;
	Qptr = players[Qptr].nextQ;
    }
    if (Qptr == id) {
	players[lastptr].nextQ = players[id].nextQ;
	return TRUE;
    }
    return FALSE;
}  /* delQ() */


/* Send data to set of players pointed to by ptr */
static void sendall(int mode, databuffer *data)
{
    int ptr;

    if (mode == GAMEMODE) ptr = gameplayers;
    else if (mode == TEXTMODE) ptr = textplayers;
    else {
	fprintf(stderr, "Bad mode passed to sendall()\n");
	return;
    }
    while (ptr != ENDLIST) {
	if (mode == GAMEMODE) {
	    if (players[ptr].connect & CLIENT) (void)senddata(ptr, data);
	    ptr = players[ptr].nextgame;
	}
	else {
	    if (players[ptr].connect & CLIENT) (void)senddata(ptr, data);
	    ptr = players[ptr].nexttext;
	}
    }
}  /* sendall() */


/* Send text to a socket */
static void sendtxt(int id, unsigned char *text)
{
    int slen;

    slen = strlen(text);

    if (players[id].connect & LOCAL) {
	fputs(text, stdout);
	fflush(stdout);
    }
#ifdef TELNET
    else if (players[id].connect & TELNET) {
	int sent, length;

	length = slen + 1;
	while (length > 0) {
	    sent = write(players[id].sckt, text, length);
	    if (sent == -1) {
		perror("write()");
		remove_player(id);
		++drops;
		break;
	    }
	    length -= sent;
#ifdef DEBUG
	    if (length) fprintf(stderr, "write(): still %d bytes to write\n",
				length);
#endif
	    text += sent;
	}
	++outpackets;
    }
#endif
    else {
	if (slen < DATASIZE) strcpy(outbuffer.data, text);
	else {
	    memcpy(outbuffer.data, text, DATASIZE - 1);
	    outbuffer.data[DATASIZE - 1] = 0;
	    slen = DATASIZE - 1;
	}
	outbuffer.length = htons(slen + 1 + HDRSIZE);
	outbuffer.pkttype = TEXTDATA;
	(void)senddata(id, &outbuffer);
    }
}  /* sendtxt() */


/* Send text to all players in GAMEMODE */
static void sendtxt_all(unsigned char *text)
{
    int ptr;

    ptr = textplayers;
    while (ptr != ENDLIST) {
	if ((players[ptr].mode & TUNED_IN) || strncmp(text, "==>", 3) != 0)
	    sendtxt(ptr, text);
	ptr = players[ptr].nexttext;
    }
}  /* sendtxt_all() */


/* Called once every second */
static void alrm(int sig)
{
    int plr;

    framespersec = frames;  /* Speed of game in frames per second */
    frames = 0;

    for (plr = textplayers; plr != ENDLIST;
      plr = players[plr].nexttext) {
	++players[plr].idle_time;
	if (players[plr].idle_time > MAX_PLAYER_IDLE) {
	    sendtxt(plr, "Maximum idle time exceeded!\n");
	    remove_player(plr);
	}
    }

    if (! time_to_start) {
	++idle_time;
	if (idle_time > MAX_IDLE_TIME) {
	    sendtxt_all("==>Game stopped due to inactivity.\n");
	    fprintf(stderr, "Game stopped due to inactivity.\n");
	    time_to_start = -1;

	    for (plr = gameplayers; plr != ENDLIST;
	      plr = players[plr].nextgame) {
		outbuffer.length = htons(HDRSIZE);
		outbuffer.pkttype = LEAVEGAME;
		if (! senddata(plr, &outbuffer)) remove_player(plr);
		else leave_gamemode(plr);
	    }
	}
	++wait_time;
	if (wait_time > MAX_WAIT) {
	    sendtxt_all(
	    "==>Timed out while waiting for game data: game stopped.\n");
	    fprintf(stderr, "Timed out while waiting for game data.\n");
	    time_to_start = -1;

	    for (plr = gameplayers; plr != ENDLIST;
	      plr = players[plr].nextgame) {
		outbuffer.length = htons(HDRSIZE);
		outbuffer.pkttype = LEAVEGAME;
		if (! senddata(plr, &outbuffer)) remove_player(plr);
		else leave_gamemode(plr);
	    }
	}
    }

    if (time_to_start > 0) {
	--time_to_start;
	if (time_to_start == 15)
	    sendtxt_all("==>Game starting in 15 seconds.\n");
	if (time_to_start == 5)
	    sendtxt_all("==>Game starting in 5 seconds.\n");
	if (time_to_start == 0) {
	    sendtxt_all("==>Game starting now.\n");
	    idle_time = 0;
	    wait_time = 0;
	}
    }
}  /* alrm() */


/* Remove a player from the system */
static void remove_player(int id)
{
    if (players[id].mode != NOTINUSE) {
	if (players[id].password[0] != '\0') {
	    if (! write_player_data(id))
		fprintf(stderr, "Error saving player information for %s\n",
			players[id].name);
	}
	--logins;
	if (players[id].mode & GAMEMODE) leave_gamemode(id);
	if (players[id].mode & TEXTMODE) {
	    --numtext;
	    deltext(id);
	}

	players[id].name[0] = '\0';

	/* Delete player from queue (or do nothing if not in the queue) */

	(void)delQ(id);
	players[id].mode = NOTINUSE;
	if (players[id].connect & LOCAL) server_id = -1;
	else close(players[id].sckt);
    }
}  /* remove_player() */


static void remove_all_players(void)
{
    int i;

    time_to_start = -1;

    sendtxt_all("Closing down.\n");
    /*outbuffer.length = htons(HDRSIZE);
    outbuffer.pkttype = GAMECLOSE;
    sendall(GAMEMODE, &outbuffer);*/
    outbuffer.length = htons(HDRSIZE);
    outbuffer.pkttype = TEXTCLOSE;
    sendall(TEXTMODE, &outbuffer);
    for (i = 0; i < MAX_LOGINS; i++)
	if (players[i].mode != NOTINUSE) remove_player(i);
}  /* remove_all_players() */


/* Function called when user presses ctrl-C or kills the server process */
static void sighandler(int sig)
{
    if (access_on) {
	close(listen_socket);
#ifdef TELNET
	close(listen_socket2);
#endif
    }
    remove_all_players();
#ifdef HOSTFILE
    remove(HOSTFILE);
#endif
    fprintf(stderr, "\nClosing down server.\n");
    exit(0);
}
/* sighandler() */


/* Setup a socket for receiving data */
static void setup_socket(void)
{
    int sockopt;
    struct sockaddr_in sin1, sin2;

    listen_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_socket < 0) {
	perror("socket()");
	exit(1);
    }
    sockopt = 1;
    if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, (char *)&sockopt,
	sizeof sockopt) < 0) {
	perror("setsockopt()");
	exit(1);
    }

    if (fcntl(listen_socket, F_SETFL, O_NDELAY)) {
	perror("fcntl()");
	exit(1);
    }

    memset((char *)&sin1, '\0', sizeof sin1);
    sin1.sin_addr.s_addr = INADDR_ANY;	/* Any address */
    sin1.sin_family = AF_INET;
    sin1.sin_port = htons((u_short)PORT_NUMBER);
    if (bind(listen_socket, (struct sockaddr *)&sin1, sizeof sin1) < 0) {
	perror("bind()");
	exit(1);
    }
    listen(listen_socket, 5);
#ifdef TELNET
    listen_socket2 = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_socket2 < 0) {
	perror("socket()");
	exit(1);
    }
    sockopt = 1;
    if (setsockopt(listen_socket2, SOL_SOCKET, SO_REUSEADDR,
		   (char *)&sockopt, sizeof sockopt) < 0) {
	perror("setsockopt()");
	exit(1);
    }

    if (fcntl(listen_socket2, F_SETFL, O_NDELAY)) {
	perror("fcntl()");
	exit(1);
    }

    memset((char *)&sin2, '\0', sizeof sin2);
    sin2.sin_addr.s_addr = INADDR_ANY;	/* Any address */
    sin2.sin_family = AF_INET;
    sin2.sin_port = htons((u_short)TELNET_PORT);
    if (bind(listen_socket2, (struct sockaddr *)&sin2, sizeof sin2) < 0) {
	perror("bind()");
	exit(1);
    }
    listen(listen_socket2, 5);
#endif
} /* setup_socket */


/* Switch off local echoing of characters */
static void echo_off(int id)
{
    databuffer *sendbuf;

    if (! (sendbuf = allocbuf(3))) {
	perror("malloc()");
	exit(1);
    }
    if (players[id].connect & LOCAL) local_echo_off();
#ifdef TELNET
    else if (players[id].connect & TELNET) {
	sendbuf->data[0] = IAC;
	sendbuf->data[1] = WILL;
	sendbuf->data[2] = TELOPT_ECHO;
	write(players[id].sckt, sendbuf->data, 3);
	++outpackets;
    }
#endif
    else {
	sendbuf->length = htons(HDRSIZE);
	sendbuf->pkttype = TEXTECHOOFF;
	(void)senddata(id, sendbuf);
    }
    free(sendbuf);
}  /* echo_off() */


/* Switch on local echoing of characters */
static void echo_on(int id)
{
    databuffer *sendbuf;

    if (! (sendbuf = allocbuf(3))) {
	perror("malloc()");
	exit(1);
    }
    if (players[id].connect & LOCAL) local_echo_on();
#ifdef TELNET
    else if (players[id].connect & TELNET) {
	sendbuf->data[0] = IAC;
	sendbuf->data[1] = WONT;
	sendbuf->data[2] = TELOPT_ECHO;
	write(players[id].sckt, sendbuf->data, 3);
	++outpackets;
    }
#endif
    else {
	sendbuf->length = htons(HDRSIZE);
	sendbuf->pkttype = TEXTECHOON;
	(void)senddata(id, sendbuf);
    }
    free(sendbuf);
}  /* echo_on() */


/* Give a player a new password */
static void new_password(int id)
{
    unsigned char buf[200];

    echo_on(id);
    sendtxt(id, "\n");
    if (inbuffer.data[0] != '\0') {
	memcpy(players[id].password, crypt(inbuffer.data, SALT), PASSLEN + 1);
	sendtxt(id, "\nAgain: ");
	input_to(id, new_password2);
	echo_off(id);
    }
    else {
#ifdef CHECK_IP
	check_ip(id);
#endif		
	cat_file(id, "LOGIN");
	sprintf(buf, "==>%s has logged in.\n", players[id].name);
	sendtxt_all(buf);
	input_to(id, NULL);
    }
}  /* new_password() */


static void new_password2(int id)
{
    unsigned char buf[200];

    echo_on(id);
    sendtxt(id, "\n");
    if (strcmp(crypt(inbuffer.data, players[id].password),
      players[id].password) != 0) {
	sendtxt(id, "Passwords different!\nPlease enter a password: ");
	input_to(id, new_password);
	echo_off(id);
    }
    else {
	input_to(id, NULL);
#ifdef CHECK_IP
	check_ip(id);
#endif		
	cat_file(id, "LOGIN");
	sprintf(buf, "==>%s has logged in.\n", players[id].name);
	sendtxt_all(buf);
    }
}  /* new_password2() */


/* Change the player's password */
static void change_password(int id)
{
    players[id].saved_scores = FALSE;
    echo_on(id);
    sendtxt(id, "\n");
    if (strcmp(crypt(inbuffer.data, players[id].password),
	       players[id].password) != 0) {
	sendtxt(id, "Wrong password.\n");
	input_to(id, NULL);
	return;
    }
    sendtxt(id, "Enter your new password: ");
    echo_off(id);
    input_to(id, change_password2);
}  /* change_password() */


static void change_password2(int id)
{
    echo_on(id);
    sendtxt(id, "\n");
    memcpy(players[id].password, crypt(inbuffer.data, SALT), PASSLEN + 1);
    sendtxt(id, "Again: ");
    echo_off(id);
    input_to(id, change_password3);
}  /* change_password2() */


static void change_password3(int id)
{
    echo_on(id);
    sendtxt(id, "\n");
    input_to(id, NULL);
    if (strcmp(crypt(inbuffer.data, players[id].password),
	       players[id].password) != 0) {
	sendtxt(id, "Passwords different!\nRestoring from save-file.\n");
	(void)read_player_data(id);
	return;
    }
    sendtxt(id, "Password changed.\n");
}  /* change_password3() */


/* Read in the player's password */
static void get_password(int id)
{
    unsigned char buf[200];

    echo_on(id);
    sendtxt(id, "\n");
    if (strcmp(crypt(inbuffer.data, players[id].password),
	       players[id].password) != 0) {
	sendtxt(id, "Wrong password.\n");
	remove_player(id);
	return;
    }
    else {
#ifdef CHECK_IP
	check_ip(id);
#endif		
	cat_file(id, "LOGIN");
	sprintf(buf, "==>%s has logged in.\n", players[id].name);
	sendtxt_all(buf);
	input_to(id, NULL);
    }
}  /* get_password() */


/* Read in a player's name */
static void get_name(int id)
{
    int slen, i;

    slen = strlen(inbuffer.data);
    if ((slen < 1) || (slen > NAMELEN))
	sendtxt(id, "Please enter your name again: ");
    else {
	if (strpbrk(inbuffer.data, wspace)) {
	    sendtxt(id,
 "Name cannot have white space characters!\nPlease enter your name again: ");
	    return;
	}
	if (strchr(inbuffer.data, ':')) {
	    sendtxt(id,
 "Name cannot have colons!\nPlease enter your name again: ");
	    return;
	}
	if (inbuffer.data[0]) inbuffer.data[0] = toupper(inbuffer.data[0]);
	strcpy(players[id].name, inbuffer.data);
	for (i = 0; i < MAX_LOGINS; i++) {
	    if ((! stricmp(players[i].name, players[id].name)) &&
		(id != i)) {
		players[id].name[0] = '\0';
		sendtxt(id,
			"Sorry, please choose a different name: ");
		return;
	    }
	}
	if (players[id].name[0] != '\0') {
	    if (read_player_data(id)) {
		sendtxt(id, "Please enter your password: ");
		echo_off(id);
		input_to(id, get_password);
	    }
	    else {
		sendtxt(id,
    "Please enter a password if you want to save your scores: ");
		echo_off(id);
		input_to(id, new_password);
	    }
	}
    }
}  /* get_name() */


/* Make connection to a client */
static void make_connection(void)
{
    int len;
    struct sockaddr_in client;
    int skt;
    int id;

    len = sizeof client;
    skt = accept(listen_socket, (struct sockaddr *)&client, &len);

    if (skt < 0) perror("accept()");
    else if (fcntl(skt, F_SETFL, O_NDELAY)) perror("fcntl()");
    else {
	id = 0;
	/* Find position in table for player */
	while ((id < MAX_LOGINS - 1) && (players[id].mode != NOTINUSE)) id++;
	if (players[id].mode == NOTINUSE) {
	    players[id].sckt = skt;
	    players[id].mode = TEXTMODE | TUNED_IN;
	    input_to(id, NULL);
	    players[id].name[0] = '\0';
	    players[id].version[0] = '\0';
	    players[id].password[0] = '\0';
	    players[id].connect = CLIENT;
	    players[id].games = players[id].wins = players[id].kills = 0;
	    players[id].saved_scores = FALSE;
	    players[id].idle_time = 0;
	    strcpy(players[id].prompt, default_prompt);
	    strncpy(players[id].ip,
		    inet_ntoa(client.sin_addr), IPLEN - 1);
	    players[id].ip[IPLEN - 1] = 0;
	    players[id].login_time = get_time();
	    ++logins;
	    ++numtext;
	    addtext(id);
	}
	else {
	    outbuffer.length = htons(HDRSIZE);
	    outbuffer.pkttype = TEXTNOLOGIN;
	    write(skt, (char *)&outbuffer, HDRSIZE);
	    ++outpackets;
	    close(skt);
	}
    }
}  /* make_connection() */


#ifdef TELNET
/* Make connection to a telnet connection */
static void make_connection_telnet(void)
{
    int len;
    struct sockaddr_in client;
    int skt;
    int id;
    char buf[4];

    len = sizeof client;
    skt = accept(listen_socket2, (struct sockaddr *)&client, &len);

    if (skt < 0) perror("accept()");
    else if (fcntl(skt, F_SETFL, O_NDELAY)) perror("fcntl()");
    else {
	id = 0;
	/* Find position in table for player */
	while ((id < MAX_LOGINS - 1) && (players[id].mode != NOTINUSE)) id++;
	if (players[id].mode == NOTINUSE) {
	    players[id].sckt = skt;
	    players[id].mode = TEXTMODE | TUNED_IN;
	    input_to(id, NULL);
	    players[id].name[0] = '\0';
	    players[id].password[0] = '\0';
	    players[id].connect = TELNET;
	    players[id].games = players[id].wins = players[id].kills = 0;
	    players[id].saved_scores = FALSE;
	    players[id].idle_time = 0;
	    strcpy(players[id].prompt, default_prompt);
	    strncpy(players[id].ip,
		    inet_ntoa(client.sin_addr), IPLEN - 1);
	    players[id].ip[IPLEN - 1] = 0;
	    players[id].login_time = get_time();
	    ++logins;
	    ++numtext;
	    addtext(id);
	    strcpy(players[id].version, version);
	    cat_file(id, "CONNECT");
	    sendtxt(id, "Please enter your name: ");
	    input_to(id, get_name);

	    /*buf[0] = IAC;
	    buf[1] = DONT;
	    buf[2] = TELOPT_LINEMODE;
	    if (write(players[id].sckt, buf, 3) != 3) {
		remove_player(id);
	    }*/
	}
	else {
	    sprintf(outbuffer.data,
		"Maximum number of logins reached (%d), try again later.\n",
		MAX_LOGINS);
	    write(skt, outbuffer.data, strlen(outbuffer.data) + 1);
	    ++outpackets;
	    close(skt);
	}
    }
}  /* make_connection_telnet() */
#endif


/* Instruct a client to parse a given file */
static void parse_file(int id, char *name)
{
    struct stat buf;
    databuffer *sendbuf;
    FILE *infile;

    if (! (infile = fopen(name, "r"))) {
	perror("fopen()");
	return;
    }
    if (stat(name, &buf)) {
	perror("stat()");
	return;
    }
    if (buf.st_size > DATASIZE - 1) {
	fprintf(stderr, "File too long.\n");
	return;
    }
    if (! (sendbuf = (databuffer *)malloc(HDRSIZE + buf.st_size + 1))) {
	perror("malloc()");
	return;
    }
    if (fread(sendbuf->data, buf.st_size, 1, infile) != 1) {
	 perror("fread()");
	 free(sendbuf);
	 return;
    }
    fclose(infile);
    sendbuf->data[buf.st_size] = '\0';  /* NULL terminator */
    sendbuf->length = htons(buf.st_size + 1 + HDRSIZE);
    sendbuf->pkttype = GAMEPARSE;
    (void)senddata(id, sendbuf);
    free(sendbuf);
}  /* parse_file() */


/* Instruct all clients to parse a given file */
static void parse_file_all(char *name)
{
    struct stat buf;
    databuffer *sendbuf;
    FILE *infile;

    if (! (infile = fopen(name, "r"))) {
	perror("fopen()");
	return;
    }
    if (stat(name, &buf)) {
	perror("stat()");
	return;
    }
    if (buf.st_size > SHRT_MAX - HDRSIZE - 1) {
	fprintf(stderr, "File too long.\n");
	return;
    }
    if (! (sendbuf = (databuffer *)malloc(HDRSIZE + buf.st_size + 1))) {
	perror("malloc()");
	return;
    }
    if (fread(sendbuf->data, buf.st_size, 1, infile) != 1) {
	 perror("fread()");
	 free(sendbuf);
	 return;
    }
    fclose(infile);
    sendbuf->data[buf.st_size] = '\0';  /* NULL terminator */
    sendbuf->length = htons(buf.st_size + 1 + HDRSIZE);
    sendbuf->pkttype = GAMEPARSE;
    sendall(GAMEMODE, sendbuf);
    free(sendbuf);
}  /* parse_file_all() */


/* Instruct a client to parse given data */
static void parse_data(int id, unsigned char *data)
{
    int slen;

    slen = strlen(data);
    if (slen > DATASIZE - 1) {
	fprintf(stderr, "Parse data too long.\n");
	return;
    }
    outbuffer.length = htons(HDRSIZE + strlen(data) + 1);
    outbuffer.pkttype = GAMEPARSE;
    strcpy(outbuffer.data, data);
    (void)senddata(id, &outbuffer);
}  /* parse_data() */


/* Instruct all clients to parse given data */
static void parse_data_all(unsigned char *data)
{
    int slen;

    slen = strlen(data);
    if (slen > DATASIZE - 1) {
	fprintf(stderr, "Parse data too long.\n");
	return;
    }
    outbuffer.length = htons(HDRSIZE + strlen(data) + 1);
    outbuffer.pkttype = GAMEPARSE;
    strcpy(outbuffer.data, data);
    sendall(GAMEMODE, &outbuffer);
}  /* parse_data_all() */


/* Instruct all clients, who are in a given world, to parse some data */
static void parse_data_all_world(unsigned char *data, char *world)
{
    int slen;
    int ptr;

    slen = strlen(data);
    if (slen > DATASIZE - 1) {
	fprintf(stderr, "Parse data too long.\n");
	return;
    }
    outbuffer.length = htons(HDRSIZE + strlen(data) + 1);
    outbuffer.pkttype = GAMEPARSE;
    strcpy(outbuffer.data, data);

    for (ptr = gameplayers; ptr != ENDLIST; ptr = players[ptr].nextgame) {
	if (strcmp(players[ptr].world, world) == 0)
	    (void)senddata(ptr, &outbuffer);
    }
}  /* parse_data_all_world() */


static void chat(int id)
{
    int slen;
    static unsigned char buf[DATASIZE];

    slen = strlen(inbuffer.data);
    if (strcmp(inbuffer.data, ".") == 0) input_to(id, NULL);
    else {
	/* Need to send the NULL char at end of string */
	sprintf(buf, "%s says: ", players[id].name);
	if (slen < (sizeof buf) - strlen(buf) - 2) {
	    sprintf(buf + strlen(buf), "%s\n", inbuffer.data);
	    sendtxt_all(buf);
	 }
	 else sendtxt(id, "Buffer overflow!\n");
	 sendtxt(id, "]");
    }
}  /* chat() */


/*
 * Checks to see if another player is using a certain colour.
 * Returns FALSE if another player is using the colour.
 */
static int check_colour(int check_colour)
{
    int ptr;

    for (ptr = gameplayers; ptr != ENDLIST; ptr = players[ptr].nextgame)
	if (players[ptr].colour == check_colour) return FALSE;

    return TRUE;
}  /* check_colour() */


/* Routine to put a player into gamemode. */
static void enter_gamemode(int id)
{
    int ptr;
    int i;
    unsigned char buf[200];

    if (players[id].mode & GAMEMODE) {
	fprintf(stderr, "Error: client already in gamemode.\n");
	return;
    }

    players[id].saved_scores = FALSE;

    /*
     * Each player entering gamemode gets allocated a number, depending
     * on whether they were the first, second, third etc. person to
     * enter the game.	This number is used to determine what colour the
     * player will be, and will also be sent with the ENTERGAME packet
     * to the client (and the client will then use it to determine
     * the starting position of the player).
     */

    players[id].colour = numgame;
    if (check_colour(numgame))
	players[id].colour = numgame;
    else {
	/*
	 * Someone else is already using this colour (this will happen
	 * if a player quits the game before it starts).
	 */
	for (i = 0; i < numgame; i++) {
	    if (check_colour(i)) {
		players[id].colour = i;
		break;
	    }
	}

	if (i == numgame) {
	    fprintf(stderr, "Error: no colour available for player.\n");
	    sendtxt(id, "Error: no colour available!\n");
	    return;
	}
    }

    outbuffer.length = htons(HDRSIZE + 1);
    outbuffer.pkttype = ENTERGAME;
    outbuffer.data[0] = players[id].colour;
    if (! senddata(id, &outbuffer)) {
	remove_player(id);
	return;
    }

    strcpy(players[id].world, WORLDFILE);   /* Starting world for player */

    /* Let other players know */

    sprintf(buf, "==>%s has entered the game (%s).\n",
	players[id].name, playercols[players[id].colour]);
    sendtxt_all(buf);

    sprintf(buf, "player \"%s\" %d \"plr_%s\" shoot;",
	    players[id].name, id,
	    playercols[players[id].colour]);
    parse_data_all_world(buf, players[id].world);

    players[id].mode |= GAMEMODE;
    if (! numgame) {
	gamecount = 0;
	gamebufptr = gamebuffer.data;
	heal_object = FALSE;
    }
    if (numgame == 1) {
	time_to_start = 30;
	sendtxt_all("==>Game starting in 30 seconds.\n");
    }
    ++numgame;
    addgame(id);
    players[id].energy = MAX_ENERGY;
    parse_file(id, WORLDDEFS);
    parse_file(id, players[id].world);

    /* Tell player who else is playing */

    for (ptr = gameplayers; ptr != ENDLIST; ptr = players[ptr].nextgame) {
	if (ptr != id &&
	  (strcmp(players[ptr].world, players[id].world) == 0)) {
	    sprintf(buf,
		"player \"%s\" %d \"plr_%s\" shoot;",
		players[ptr].name, ptr,
		playercols[players[ptr].colour]);
	    parse_data(id, buf);
	}
    }
}  /* enter_gamemode() */


/* This routine is called when a player leaves game mode.
   It should be called after the client has been sent a LEAVEGAME packet
   (or after the server has received one). */
static void leave_gamemode(int id)
{
    unsigned char buf[200];

    if (! (players[id].mode & GAMEMODE)) {
	fprintf(stderr, "Error: client not in gamemode.\n");
	return;
    }

    check_high_score(&players[id]);

    if (time_to_start == 0) ++players[id].games;
    players[id].mode ^= GAMEMODE;
    --numgame;

    /* Remove from game list */

    delgame(id);

    /* Tell other players */

    sprintf(buf, "destroy \"%s\";", players[id].name);
    parse_data_all_world(buf, players[id].world);
    if (numgame == 1) {

	/* Only one other person playing */

	if (time_to_start == 0) {
	    games++;
	    sprintf(buf, "==>%s has won the game.\n",
	      players[gameplayers].name);
	    sendtxt_all(buf);
	    ++players[gameplayers].wins;
	}
	outbuffer.length = htons(HDRSIZE);
	outbuffer.pkttype = LEAVEGAME;
	if (! senddata(gameplayers, &outbuffer)) remove_player(gameplayers);
	else leave_gamemode(gameplayers);
    }
    if ((gamecount >= numgame) && (numgame)) {

	/* Send data to all clients playing game */

	++frames;
	gamebuffer.length = htons(gamebufptr - (unsigned char *)&gamebuffer);
	gamebuffer.pkttype = GAMEDATA;
	sendall(GAMEMODE, &gamebuffer);
	gamebufptr = (unsigned char *)&gamebuffer + HDRSIZE;
	gamecount = 0;
    }
    if (numgame == 0) {
	time_to_start = -1;
	if (allow_games) while (queue != ENDLIST && numgame < MAX_PLAYERS) {
	    enter_gamemode(queue);

	    /* Remove player from queue */

	    queue = players[queue].nextQ;
	}
    }
}  /* leave_gamemode() */


/* Routine to parse command in inbuffer.data */
static void parse_input(int id)
{
    unsigned char cmdbuffer[255];
    unsigned char arg1[255];
    static unsigned char buf[DATASIZE];
    char *ptr;

    int i;

    /* First remove all control characters from input string */

    for (ptr = inbuffer.data; *ptr != '\0'; ptr++)
	if (*ptr < 32 || *ptr > 126) *ptr = ' ';

    if (players[id].version[0] == '\0') {
	/* Client hasn't sent version yet */
	int slen;

	slen = strlen(inbuffer.data);
	if ((slen < 1) || (slen > TYPELEN)) remove_player(id);
	else {
	    strcpy(players[id].version, inbuffer.data);
	    /* First send id to player, then login string */
	    outbuffer.length = htons(HDRSIZE + 1);
	    outbuffer.pkttype = TEXTLOGIN;
	    outbuffer.data[0] = id;
	    ++outpackets;
	    if (strcmp(version, players[id].version)) {
		sendtxt(id, "Wrong version of client!\n");
		remove_player(id);
		return;
	    }
	    if (! senddata(id, &outbuffer)) {
		perror("Error sending initial data to client");
		remove_player(id);
	    }
	    else {
		cat_file(id, "CONNECT");
		sendtxt(id, "Please enter your name: ");
		input_to(id, get_name);
	    }
	}
    }
    else if (players[id].func) (*players[id].func)(id);
    else {
	sscanf(inbuffer.data, "%80s", cmdbuffer);
	if (inbuffer.data[0]) {
	    if (! strcmp(cmdbuffer, "quit")) {
		sendtxt(id, "Closing connection.\n");
		if (players[id].mode & GAMEMODE) {
		    outbuffer.length = htons(HDRSIZE);
		    outbuffer.pkttype = LEAVEGAME;
		    (void)senddata(id, &outbuffer);
		    leave_gamemode(id);
		}
		outbuffer.length = htons(HDRSIZE);
		outbuffer.pkttype = TEXTCLOSE;
		(void)senddata(id, &outbuffer);
		sprintf(buf,
		  "==>%s has logged out.\n", players[id].name);
		remove_player(id);
		sendtxt_all(buf);
	    }
	    else if (! strcmp(cmdbuffer, "leave")) {
		if (! (players[id].mode & GAMEMODE)) {
		    if (delQ(id)) {
			sendtxt(id,
			  "You have been removed from the waiting list.\n");
		    }
		    else sendtxt(id, "Not in game mode!\n");
		}
		else {
		    outbuffer.length = htons(HDRSIZE);
		    outbuffer.pkttype = LEAVEGAME;
		    if (! senddata(id, &outbuffer)) remove_player(id);
		    else leave_gamemode(id);
		    sprintf(buf,
		      "==>%s has quit the game.\n", players[id].name);
		    sendtxt_all(buf);
		}
	    }
	    else if (! strcmp(cmdbuffer, "start")) {
		if (! allow_games)
		    sendtxt(id, "Game mode is currently disabled.\n");
		else if (! (players[id].connect & CLIENT))
		    sendtxt(id, "Can't start game on a text terminal!\n");
		else if (players[id].mode & GAMEMODE)
		    sendtxt(id, "Already in game mode!\n");
		else if (numgame >= MAX_PLAYERS) {
		    int posn;

		    posn = addQ(id);
		    if (posn) sprintf(buf,
   "Too many players already playing; adding to waiting list (position %d).\n",
   posn);
		    else sprintf(buf,
    "Be patient, you're already on the waiting list!\n");
		    sendtxt(id, buf);
		}
		else if (numgame > 0 && time_to_start == 0) {
		    int posn;

		    posn = addQ(id);
		    if (posn) sprintf(buf,
   "Game already in progress; adding to waiting list (position %d).\n",
   posn);
		    else sprintf(buf,
    "Be patient, you're already on the waiting list!\n");
		    sendtxt(id, buf);
		}
		else enter_gamemode(id);
	    }
	    else if (! strcmp(cmdbuffer, "save")) {
		if (players[id].password[0] == '\0')
		    sendtxt(id, "You don't have a password!\n");
		else if (players[id].saved_scores)
		    sendtxt(id, "No change since last save!\n");
		else if (! write_player_data(id)) {
		    fprintf(stderr,
		      "Error saving player information for %s\n",
		      players[id].name);
		    sendtxt(id, "Error while saving your scores.\n");
		}
		else sendtxt(id, "Scores saved.\n");
	    }
	    else if (! strcmp(inbuffer.data, "tune out")) {
		sendtxt(id, "Game information messages will not be sent.\n");
		players[id].mode &= ~TUNED_IN;
	    }
	    else if (! strcmp(inbuffer.data, "tune in")) {
		sendtxt(id, "Game information messages will be sent.\n");
		players[id].mode |= TUNED_IN;
	    }
	    else if (! strcmp(cmdbuffer, "who")) {
		if (logins == 1) sprintf(buf,
				   "You are the only player logged on.\n");
		else sprintf(buf, "%d players logged on.\n", logins);
		strcat2(buf,
"\nPlayer              Status  Logged on from   Games  Wins  Kills  Colour\n");
		strcat2(NULL,
  "=======================================================================\n");
		for (i = 0; i < MAX_LOGINS; i++) {
		    if (players[i].mode != NOTINUSE) {
			sprintf(buf + strlen(buf), "%-20s", players[i].name);
			if (players[i].mode & TEXTMODE) strcat2(NULL, "T");
			else strcat2(NULL, "-");
			if (players[i].mode & GAMEMODE) strcat2(NULL, "G");
			else strcat2(NULL, "-");
			if (players[i].mode & HACKMODE) strcat2(NULL, "A");
			else strcat2(NULL, "-");
			strcat2(NULL, "     ");
			sprintf(buf + strlen(buf), "%-17s%-7d%-6d%-7d",
				players[i].ip, players[i].games,
				players[i].wins,
				players[i].kills);
			if (players[i].mode & GAMEMODE)
			    strcpy(buf + strlen(buf),
				   playercols[players[i].colour]);
			strcat(buf + strlen(buf), "\n");
		    }
		}
		strcat(buf, "(Status: [T]extmode, [G]amemode, [A]dmin)\n");
		sendtxt(id, buf);
	    }

	    else if (! strcmp(cmdbuffer, "password")) {
		if (players[id].password[0] == '\0')
		    sendtxt(id, "You don't have a password!\n");
		else {
		    sendtxt(id, "Enter your old password: ");
		    input_to(id, change_password);
		    echo_off(id);
		}
	    }
	    else if ((sscanf(inbuffer.data, "tell %80s", arg1)) == 1) {
		int plr;

		if ((plr = find_player(arg1)) >= 0) {
		    sprintf(buf, "%s tells you: ", players[id].name);
		    if (strlen(inbuffer.data) < (sizeof buf)
		      - strlen(buf) - 2) {
			sscanf(inbuffer.data, "tell %*s %[^\n]",
			       buf + strlen(buf));
			sprintf(buf + strlen(buf), "\n");

			sendtxt(plr, buf);
			sprintf(buf, "You tell %s: ", players[plr].name);
			sscanf(inbuffer.data, "tell %*s %[^\n]",
			       buf + strlen(buf));
			sprintf(buf + strlen(buf), "\n");
			sendtxt(id, buf);
		    }
		    else sendtxt(id, "Buffer overflow!\n");
		}
		else {
		    sprintf(buf, "Cant find player '%s'.\n", arg1);
		    sendtxt(id, buf);
		}
	    }
	    else if (! strcmp(cmdbuffer, "status")) {
		long packets_per_sec;
		float bwidth;	 /* % of bandwidth */

		/*
		 * Calculate the average number of GAMEDATA packets sent and
		 * received each second, and then calculate the ethernet
		 * bandwidth used.
		 */

		packets_per_sec = numgame * numgame * framespersec;

		bwidth = packets_per_sec * PKTSIZE_AVERAGE * 8.0 /
			 ETHER_SPEED * 100.0;

		strcpy(buf, "Server up since: ");
		print_time(up_since, buf + strlen(buf));
		strcat(buf, "\n");
		sprintf(buf + strlen(buf),
     "Games played since server started: %ld\n", games);
		sprintf(buf + strlen(buf),
     "%d players connected (%d in text mode and %d in game mode).\n",
			logins, numtext, numgame);
		sprintf(buf + strlen(buf),
     "Total tcp packets in: %ld\nTotal tcp packets out: %ld\n",
			inpackets, outpackets);
		sprintf(buf + strlen(buf),
			"Dropped packets: %ld\n", drops);
		sprintf(buf + strlen(buf),
			"Maximum logins: %d\n", MAX_LOGINS);
		sprintf(buf + strlen(buf),
			"Maximum players in game: %d\n", MAX_PLAYERS);
		if (access_on) {
		    sprintf(buf + strlen(buf),
			    "Listening for client connections on port %d\n",
			    PORT_NUMBER);
#ifdef TELNET
		    sprintf(buf + strlen(buf),
			    "Listening for telnet connections on port %d\n",
			    TELNET_PORT);
#endif
		}
		else strcat(buf, "Not accepting new connections.\n");
		if (numgame) {
		    sprintf(buf + strlen(buf),
			"Update information being sent %d times per second.\n",
			framespersec);
		    sprintf(buf + strlen(buf),
			"Estimated ethernet bandwidth usage: %.3f%%\n", bwidth);
		}
		sendtxt(id, buf);
	    }
	    else if (! strcmp(cmdbuffer, "stop")) {
		int plr;

		if (players[id].mode & HACKMODE) {
		    if (time_to_start)
			sendtxt(id, "No game in progress!\n");
		    else {
			int plr = gameplayers;

			time_to_start = -1;
			while (plr != ENDLIST) {
			    outbuffer.length = htons(HDRSIZE);
			    outbuffer.pkttype = LEAVEGAME;
			    if (! senddata(plr, &outbuffer)) remove_player(plr);
			    else leave_gamemode(plr);

			    plr = players[plr].nextgame;
			}
			queue = ENDLIST;
			sendtxt_all("==>Everyone has quit the game.\n");
		    }
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "parse")) {
		int slen;

		if (players[id].mode & HACKMODE) {
		    slen = strlen(inbuffer.data);
		    if (slen > 6)
			parse_data_all(&inbuffer.data[6]);
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "parsefile")) {
		int slen;

		if (players[id].mode & HACKMODE) {
		    slen = strlen(inbuffer.data);
		    if (slen > 10)
			parse_file_all(&inbuffer.data[10]);
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if ((sscanf(inbuffer.data, "help %20s", arg1)) == 1) {
		sprintf(buf, HELPDIR);
		strcat2(buf, "/");
		strcat2(NULL, arg1);
		more_file(id, buf);
	    }
	    else if (! strcmp(cmdbuffer, "help")) {
		sprintf(buf, HELPDIR);
		strcat(buf, "/commands");
		cat_file(id, buf);
		if (players[id].mode & HACKMODE) {
		    sprintf(buf, HELPDIR);
		    strcat(buf, "/extra_commands");
		    cat_file(id, buf);
		}
	    }
	    else if (! strcmp(cmdbuffer, "shutdown")) {
		if (players[id].mode & HACKMODE) {
		    remove_all_players();
		    shutdownmode = TRUE;
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "access_off")) {
		if (players[id].mode & HACKMODE) {
		    if (access_on) {
			sendtxt(id, "Turning off TCP/IP login access.\n");
			access_on = FALSE;
			close(listen_socket);
#ifdef TELNET
			close(listen_socket2);
#endif
		    }
		    else sendtxt(id, "Access is already turned off!\n");
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if ((sscanf(inbuffer.data, "remove %80s", arg1)) == 1) {
		int plr;

		if (players[id].mode & HACKMODE) {
		    if ((plr = find_player(arg1)) >= 0) {
			if (players[plr].mode & TEXTMODE) {
			    sendtxt(plr,
				"You have been removed from the game.\n");
			    outbuffer.length = htons(HDRSIZE);
			    outbuffer.pkttype = TEXTCLOSE;
			    (void)senddata(plr, &outbuffer);
			}
			else {
			    outbuffer.length = htons(HDRSIZE);
			    outbuffer.pkttype = GAMECLOSE;
			    (void)senddata(plr, &outbuffer);
			}
			remove_player(plr);

			sprintf(buf, "%s has been removed from the game.\n",
				arg1);
			sendtxt_all(buf);
		    }
		    else {
			sprintf(buf, "Cant find player '%s'.\n", arg1);
			sendtxt(id, buf);
		    }
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if ((sscanf(inbuffer.data, "purge %80s", arg1)) == 1) {
		if (players[id].mode & HACKMODE) {
		    if (atoi(arg1) < 1) sendtxt(id, "Months must be > 0\n");
		    else if (purge(atoi(arg1)))
			sendtxt(id, "Purge successful.\n");
		    else sendtxt(id, "Purge unsuccessful.\n");
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "access_on")) {
		if (players[id].mode & HACKMODE) {
		    if (! access_on) {
			sendtxt(id, "Turning on TCP/IP login access.\n");
			access_on = TRUE;
			setup_socket();
		    }
		    else sendtxt(id, "Access is already turned on!\n");
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "allow")) {
		if (players[id].mode & HACKMODE) {
		    if (! allow_games) {
			sendtxt(id, "Allowing games.\n");
			allow_games = TRUE;
		    }
		    else sendtxt(id, "Games are already allowed!\n");
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "disallow")) {
		if (players[id].mode & HACKMODE) {
		    if (allow_games) {
			sendtxt(id, "Disallowing games.\n");
			allow_games = FALSE;
			queue = ENDLIST;
		    }
		    else sendtxt(id, "Games are already disallowed!\n");
		}
		else sendtxt(id, "Command not available.\n");
	    }
	    else if (! strcmp(cmdbuffer, "chat")) {
		sendtxt(id, "Type '.' on a line to finish.\n");
		input_to(id, chat);
	    }
	    else if (! strcmp(cmdbuffer, "show")) {
		int queueptr, i;

		sprintf(buf, "Current waiting list:\n");
		for (queueptr = queue, i = 1; queueptr != ENDLIST;
		     queueptr = players[queueptr].nextQ, i++) {
		    sprintf(buf + strlen(buf), "%3d: %s\n", i,
			    players[queueptr].name);
		}
		sendtxt(id, buf);
	    }
	    else if (! strcmp(cmdbuffer, "best")) {
		int i;
		float wins_per_game, kills_per_game;

		sprintf(buf,  "Top players:\n");
		strcat2(buf,
  "Rank      Player             Kills     Kills/Game     Wins/Game\n");
		strcat2(NULL,
  "===============================================================\n");
		for (i = HIGH_SCORES - 1; i >= 0; i--) {
		    if (high_scores[i].games == 0)
			wins_per_game = kills_per_game = 0.0;
		    else {
			wins_per_game = high_scores[i].wins /
					(float)high_scores[i].games;
			kills_per_game = high_scores[i].kills /
					(float)high_scores[i].games;
		    }
		    sprintf(buf + strlen(buf),
			"%-10d%-19s%-10d%-15.2f%-6.2f\n", HIGH_SCORES - i,
			high_scores[i].name, high_scores[i].kills,
			kills_per_game, wins_per_game);
		}
		sendtxt(id, buf);
	    }
	    else if (! strcmp(cmdbuffer, "say")) {
		int slen;

		slen = strlen(inbuffer.data);
		if (slen > 4) {
		    /* Need to send the NULL char at end of string */
		    sprintf(buf, "%s says: ", players[id].name);
		    if (slen < (sizeof buf) - strlen(buf) - 2) {
			sprintf(buf + strlen(buf), "%s\n", &inbuffer.data[4]);
			sendtxt_all(buf);
		    }
		    else sendtxt(id, "Buffer overflow!\n");
		}
	    }
	    else if (inbuffer.data[0] ==  '\'') {
		int slen;

		slen = strlen(inbuffer.data);
		if (slen > 1) {
		    /* Need to send the NULL char at end of string */
		    sprintf(buf, "%s says: ", players[id].name);
		    if (slen < (sizeof buf) - strlen(buf) - 1) {
			sprintf(buf + strlen(buf), "%s\n", &inbuffer.data[1]);
			sendtxt_all(buf);
		    }
		    else sendtxt(id, "Buffer overflow!\n");
		}
	    }
	    else if (! strcmp(cmdbuffer, "emote")) {
		int slen;

		slen = strlen(inbuffer.data);
		if (slen > 6) {
		    /* Need to send the NULL char at end of string */
		    sprintf(buf, "%s ", players[id].name);
		    if (slen < (sizeof buf) - strlen(buf) - 2) {
			sprintf(buf + strlen(buf), "%s\n", &inbuffer.data[6]);
			sendtxt_all(buf);
		    }
		    else sendtxt(id, "Buffer overflow!\n");
		}
	    }
	    else sendtxt(id, "Unknown command.\n");
	}
    }
    if ((players[id].name[0] != '\0') && (! shutdownmode) &&
      (! players[id].func))
	sendtxt(id, players[id].prompt);
}  /* parse_input() */


#ifdef TELNET
/* Remove telnet protocol sequences, responding with "I won't do that!" */
/* also strip CRs and LFs */
static int strip_telnet_commands(unsigned char *to, unsigned char *from,
    int bytes, int id)
{
    unsigned char buf[3];
    unsigned char *start = to;

    while (bytes-- > 0) {
	if ((*from == 10) || (*from == 13)) {
	    from++;
	    *to++ = 0;
	}
	else if (*from != IAC) *to++ = *from++;
	else {
	    ++from;
	    bytes -= 2;
	    if (*from++ == DO && *from != TELOPT_ECHO &&
	      *from != TELOPT_LINEMODE) {
		buf[0] = IAC;
		buf[1] = WILL;
		buf[2] = *from;
		if (write(players[id].sckt, buf, 3) != 3) {
		    remove_player(id);
		    break;
		}
	    }
	    ++from;
	}
    }
    return (int)(to - start);  /* Actual number of characters sent */
}  /* strip_telnet_commands() */
#endif


/* Check a player's energy */
static void check_energy(
	int id,     /* Player doing the shooting */
	int plr)    /* Player who has been shot */
{
    unsigned char buf[200];

    if (players[plr].mode & GAMEMODE) {
	if (players[plr].energy < 0) {
	    sprintf(buf, "==>%s has been killed by %s.\n", players[plr].name,
		    players[id].name);
	    sendtxt_all(buf);
	    ++players[id].kills;
	    outbuffer.length = htons(HDRSIZE);
	    outbuffer.pkttype = LEAVEGAME;
	    if (! senddata(plr, &outbuffer)) remove_player(plr);
	    else leave_gamemode(plr);
	}
	else if (players[plr].energy % WARN == 0) {
	    sprintf(buf, "Energy at %2.0f%%\n",
		(float)players[plr].energy / MAX_ENERGY
		* 100.0);
	    sendtxt(plr, buf);
	}
    }
}  /* check_energy() */


/* Read input from a socket */
static void doio(int id)
{
    int pktlen, length;
    int plr;
    static unsigned char buf[DATASIZE];
    char *readptr;
    int bytes_to_read;
    char worldbuf[200];
    int ptr;

    players[id].idle_time = 0;

    if (players[id].connect & LOCAL) {
	fgets(inbuffer.data, DATASIZE, stdin);
	pktlen = strlen(inbuffer.data);
	if ((pktlen > 0) && (inbuffer.data[pktlen - 1] == '\n'))
	    inbuffer.data[pktlen - 1] = 0;
	parse_input(id);
    }
#ifdef TELNET
    else if (players[id].connect & TELNET) {
	pktlen = read(players[id].sckt, inbuffer.data, DATASIZE - 1);
	++inpackets;
	if (strip_telnet_commands(inbuffer.data, inbuffer.data, pktlen, id) > 0)
	    parse_input(id);
    }
#endif
    else do {
	pktlen = read(players[id].sckt, (char *)&inbuffer.length,
	  sizeof inbuffer.length);
	if (pktlen != sizeof inbuffer.length) {
	    if (pktlen == 0) {
#ifdef DEBUG
		fprintf(stderr, "Tried to read from closed socket\n");
#endif
		remove_player(id);
		return;
	    }
	    if (pktlen < 0) perror("read()");
	    ++drops;
	    fprintf(stderr, "Error receiving length: player removed\n");
	    remove_player(id);
	}
	else {
	    length = ntohs(inbuffer.length) - sizeof inbuffer.length;
	    readptr = (char *)&inbuffer + sizeof inbuffer.length;

	    while (length > 0) {
		pktlen = read(players[id].sckt, readptr, length);
		if (pktlen <= 0) {
		    perror("read()");
		    ++drops;
		    remove_player(id);
		    return;
		}
		length -= pktlen;
#ifdef DEBUG
		if (length) fprintf(stderr, "read(): still %d bytes to read\n",
				    length);
#endif
		readptr += pktlen;
	    }
	    ++inpackets;
	    if ((players[id].mode & TEXTMODE) &&
	      (inbuffer.pkttype &  TEXTPACKET)) {
		switch (inbuffer.pkttype) {
		    case TEXTDATA :
			parse_input(id);
			break;
		    case TEXTCLOSE :
			remove_player(id);
			break;
		    default :
			/* Unknown packet type */
#ifdef DEBUG
			fprintf(stderr, "Unknown packet type received\n");
#endif
			++drops;
		}
	    }
	    if (((players[id].mode & GAMEMODE) ||
		 (inbuffer.pkttype == GAMEPARSE))
	      && (inbuffer.pkttype & GAMEPACKET)) {
		switch (inbuffer.pkttype) {
		    case GAMEDATA :
			wait_time = 0;

			/* Gamedata from all clients in gamemode is
			   concatenated and sent out to each client */

			memcpy(gamebufptr, inbuffer.data,
			       ntohs(inbuffer.length) - HDRSIZE);
#if DEBUG >= 3
			fprintf(stderr, "Packet %u received from: %u\n",
				inbuffer.pkttype,
				((update *)inbuffer.data)->id);
#endif
			gamebufptr += ntohs(inbuffer.length) - HDRSIZE;
			/* Increment <gamecount> if update information is
			   for a player */
			if (((update *)inbuffer.data)->id < MAX_LOGINS)
			    ++gamecount;
			if (gamecount >= numgame) {
			    /* Send data to all clients playing game */
			    ++frames;
			    gamebuffer.length = htons(gamebufptr -
					 (unsigned char *)&gamebuffer);
			    gamebuffer.pkttype = GAMEDATA;
			    sendall(GAMEMODE, &gamebuffer);
			    gamebufptr = (unsigned char *)&gamebuffer +
					 HDRSIZE;
			    gamecount = 0;
			}
			break;
		    case GAMEWHO :
			strcpy(buf, "Players in game:\n");
			for (ptr = gameplayers; ptr != ENDLIST;
			  ptr = players[ptr].nextgame) {
			    sprintf(buf + strlen(buf), "%s (%s)\n",
				    players[ptr].name,
				    playercols[players[ptr].colour]);
			}
			sendtxt(id, buf);
			break;
		    case LEAVEGAME :
			leave_gamemode(id);
			sprintf(buf,
			    "==>%s has quit the game.\n", players[id].name);
			sendtxt_all(buf);
			break;
		    case GAMECLOSE :
			remove_player(id);
			break;
		    case GAMEPARSE :
			parse_data_all_world(inbuffer.data, players[id].world);
			break;
		    case GAMEPORTAL :
			parse_data(id, "destroy all;");
			parse_file(id, inbuffer.data);
			if ((! time_to_start) && heal_object &&
			  (strcmp(inbuffer.data, WORLDFILE) == 0))
			    parse_data(id, "create \"heal_object\";");
			strcpy(worldbuf, players[id].world);
			players[id].world[0] = '\0';
			sprintf(buf, "destroy \"%s\";", players[id].name);
			parse_data_all_world(buf, worldbuf);

			/* Tell other players */

			sprintf(buf, "player \"%s\" %d \"plr_%s\" shoot;",
				players[id].name, id,
				playercols[players[id].colour]);
			parse_data_all_world(buf, inbuffer.data);

			strcpy(players[id].world, inbuffer.data);

			/* Tell player who else is playing */

			for (ptr = gameplayers; ptr != ENDLIST;
			  ptr = players[ptr].nextgame) {
			    if (ptr != id &&
			      (strcmp(players[ptr].world,
				   players[id].world) == 0)) {
				sprintf(buf,
				    "player \"%s\" %d \"plr_%s\" shoot;",
				    players[ptr].name, ptr,
				    playercols[players[ptr].colour]);
				parse_data(id, buf);
			    }
			}

			break;
		    case EXPLOSION :
			idle_time = 0;
			if (time_to_start) break;  /* Game hasn't started */
			plr = inbuffer.data[0];
			outbuffer.pkttype = ENERGYLOSS;
			outbuffer.length = htons(HDRSIZE + 1);
			outbuffer.data[0] = BLAST_DAMAGE;
			players[plr].energy -= BLAST_DAMAGE;

			if (! senddata(plr, &outbuffer)) {
			    remove_player(plr);
			    return;
			}
			check_energy(id, plr);
			break;
		    case GAMECOLLIDE :
			if (time_to_start) break;  /* Game hasn't started */
			if (heal_object &&
			  strcmp(inbuffer.data, "heal_object") == 0) {
			    heal_object = FALSE;
			    parse_data_all_world(
			       "destroy \"heal_object\";", WORLDFILE);

			    outbuffer.pkttype = ENERGYGAIN;
			    outbuffer.length = htons(HDRSIZE + 1);
			    outbuffer.data[0] = min(HEAL_ENERGY,
				MAX_ENERGY - players[id].energy);
			    players[id].energy += min(HEAL_ENERGY,
				MAX_ENERGY - players[id].energy);
			    if (! senddata(id, &outbuffer)) {
				remove_player(id);
				return;
			    }
			}
			break;
		    case GAMESHOOT :
			idle_time = 0;
			if (time_to_start) break;
			plr = find_player(inbuffer.data);
			if (plr >= 0 && (players[plr].mode & GAMEMODE)) {
			    outbuffer.pkttype = ENERGYLOSS;
			    outbuffer.length = htons(HDRSIZE + 1);
			    outbuffer.data[0] = 1;
			    if (! senddata(plr, &outbuffer)) {
				remove_player(plr);
				return;
			    }
			    players[plr].energy -= 1;
			    check_energy(id, plr);
			}
			break;
		    default :
			/* Unknown packet type */
			++drops;
#ifdef DEBUG
			fprintf(stderr, "Unknown packet type received\n");
#endif
			break;
		}
	    }
	}

	/* See if there's anything else to read from the socket */

	if (players[id].mode != NOTINUSE) {
	    if (ioctl(players[id].sckt, FIONREAD, &bytes_to_read))
	    {
		remove_player(id);
		perror("ioctl()");
		return;
	    }
	}
    } while (players[id].mode != NOTINUSE && bytes_to_read > 0);
}  /* doio() */


static void usage(char *name)
{
	fprintf(stderr, "Usage:\n\t%s [-bg]\n", name);
}  /* usage() */


/* See if it's time to generate a heal_object */
static void check_heal(void)
{
    if ((! time_to_start) && heal_object &&
      (rand() % HEAL_DESTROY_CHANCE) == 0) {
	heal_object = FALSE;
	parse_data_all_world("destroy \"heal_object\";", WORLDFILE);
    }
    else if ((! time_to_start) && ((rand() % HEAL_CHANCE) == 0) &&
      (! heal_object)) {
	heal_object = TRUE;
	parse_data_all_world("create \"heal_object\";", WORLDFILE);
    }
}  /* check_heal() */


int main(int argc, char *argv[])
{
    struct itimerval itime, itime2;
    char cmdbuf[100];
    int i;
    fd_set readfds;
    int nds;		/* Max number of file descriptors (for select() ) */
    int iostat; 	/* Number of input descriptors ready for reading */
    int background;	/* Run server in background? */
    pid_t pid;
    sigset_t alrm_sig;

    time_to_start = -1;
    games = 0;
    shutdownmode = FALSE;
    gamebufptr = gamebuffer.data;
    gamecount = 0;
    server_id = -1;
    allow_games = TRUE;
    access_on = TRUE;  /* Allow logins */
    inpackets = outpackets = drops = 0;
    logins = numtext = numgame = 0;
    gameplayers = textplayers = queue = ENDLIST;
    up_since = get_time();

    assert(sizeof(playercols) / sizeof(*playercols) >= MAX_PLAYERS);
    assert(strlen(HELPDIR) <= FILENAMELEN);
    assert(strlen(WORLDFILE) <= FILENAMELEN);

#ifdef HOSTFILE
    fprintf(stderr, "Creating host file '%s'\n", HOSTFILE);
    sprintf(cmdbuf, "hostname > %s", HOSTFILE);
    system(cmdbuf);
    chmod(HOSTFILE, 0644);
#endif

    sigemptyset(&alrm_sig);
    sigaddset(&alrm_sig, SIGALRM);

    background = FALSE;

    if (argc > 2) {
	usage(argv[0]);
	exit (1);
    }
    if (argc == 2) {
	if (strcmp(argv[1], "-bg") == 0) background = TRUE;
	else {
	    usage(argv[0]);
	    exit(1);
	}
    }

    if (background) {
	pid = fork();
	fprintf(stderr, "Running server in background.\n");
	if (pid) exit(0);
    }

    nds = getdtablesize();
    for (i = 0; i < MAX_LOGINS; i++) players[i].mode = NOTINUSE;
    frames = 0;
    framespersec = 0;

    /* Set up a signal handler to catch ctrl-C and 'kill's */

    signal(SIGINT, sighandler);
    signal(SIGTERM, sighandler);
    signal(SIGPIPE, SIG_IGN);

    read_high_scores();

    setup_socket();

    /* Set up alarm signal to occur every second */

    itime.it_interval.tv_sec = 1;
    itime.it_interval.tv_usec = 0;
    itime.it_value.tv_sec = 1;
    itime.it_value.tv_usec = 0;
    signal(SIGALRM, alrm);
    if (setitimer(ITIMER_REAL, &itime, &itime2)) {
	fprintf(stderr, "Error setting interval timer.\n");
	exit(1);
    }

    fprintf(stderr, "Waiting for connections...\n");

    do {
	/* First set up an I/O descriptor set so we can use select() to wait
	   for either input from a player or a new connection */

	int ptr;

	FD_ZERO(&readfds);
	ptr = gameplayers;
	while (ptr != ENDLIST) {
	    FD_SET(players[ptr].sckt, &readfds);
	    ptr = players[ptr].nextgame;
	}
	ptr = textplayers;
	while (ptr != ENDLIST) {
	    FD_SET(players[ptr].sckt, &readfds);
	    ptr = players[ptr].nexttext;
	}
	if (access_on) {
	    FD_SET(listen_socket, &readfds);
#ifdef TELNET
	    FD_SET(listen_socket2, &readfds);
#endif
	}
	if (server_id < 0 && (! background)) FD_SET(fileno(stdin), &readfds);

	/* Only allow SIGALRM signals to occur during a select() */

	sigprocmask(SIG_UNBLOCK, &alrm_sig, NULL);
	iostat = select(nds, &readfds, NULL, NULL, NULL);
	sigprocmask(SIG_BLOCK, &alrm_sig, NULL);

	if (iostat < 0 && errno != EINTR) perror("select()");
	else if (iostat > 0) {
	    /* There is some input to be read */
	    ptr = gameplayers;
	    while (ptr != ENDLIST) {
		if FD_ISSET(players[ptr].sckt, &readfds) {
		    doio(ptr);
		    /* Next line makes sure a socket isn't inspected twice
		       (if a player changes from GAMEMODE to TEXTMODE) */
		    FD_CLR(players[ptr].sckt, &readfds);
		}
		ptr = players[ptr].nextgame;
	    }
	    ptr = textplayers;
	    while ((ptr != ENDLIST) && (! shutdownmode)) {
		if FD_ISSET(players[ptr].sckt, &readfds) {
		    doio(ptr);
		    FD_CLR(players[ptr].sckt, &readfds);
		}
		ptr = players[ptr].nexttext;
	    }
	    if ((access_on) && (! shutdownmode)) {
		if (FD_ISSET(listen_socket, &readfds)) make_connection();
#ifdef TELNET
		if (FD_ISSET(listen_socket2, &readfds))
		    make_connection_telnet();
#endif
	    }
	    if ((server_id < 0) && (! shutdownmode) && (! background))
	      if (FD_ISSET(fileno(stdin), &readfds)) {
		fgets(inbuffer.data, DATASIZE, stdin);
		if (strcmp(inbuffer.data, "login\n")) {
		    if (strlen(inbuffer.data) > 1) printf(
   "Unknown command - type 'login' to enter command mode.\n");
		}
		else {
		    server_id = 0;
		    /* Find position in table for player */
		    while ((server_id < MAX_LOGINS - 1) &&
			   (players[server_id].mode != NOTINUSE))
			server_id++;
		    if (players[server_id].mode == NOTINUSE) {
			players[server_id].sckt = fileno(stdin);
			players[server_id].mode = TEXTMODE | HACKMODE |
						  TUNED_IN;
			input_to(server_id, NULL);
			players[server_id].name[0] = '\0';
			players[server_id].password[0] = '\0';
			strcpy(players[server_id].version, version);
			players[server_id].connect = LOCAL;
			players[server_id].saved_scores = FALSE;
			players[server_id].idle_time = 0;
			strcpy(players[server_id].prompt, default_prompt);
			strcpy(players[server_id].ip, "LOCAL");
			players[server_id].login_time = get_time();
			players[server_id].games = 0;
			players[server_id].wins = 0;
			players[server_id].kills = 0;
			++logins;
			++numtext;
			/* Add player to linked list */
			addtext(server_id);
			cat_file(server_id, "CONNECT");
			sendtxt(server_id, "Please enter your name: ");
			input_to(server_id, get_name);
		    }
		}
	    }
	}

	check_heal();
    } while (! shutdownmode);

    if (access_on) {
	close(listen_socket);
#ifdef TELNET
	close(listen_socket2);
#endif
    }
#ifdef HOSTFILE
    remove(HOSTFILE);
#endif
    fprintf(stderr, "\nClosing down server.\n");

    return 0;
}  /* main() */

